package node

import (
	"bytes"
	"encoding/xml"
	"flag"
	"io"
	"io/ioutil"
	"log"
	"net"
	"net/url"
	"os"
	"strconv"

	"git.autistici.org/ale/autoradio"
)

var (
	icecastAdminPwFile = "/etc/icecast2/.admin_pw"
	icecastAdminPw     string

	// Icecast tunables.
	icecastQueueSize     = flag.Int("icecast-queue-size", 1<<20, "Icecast queue size (bytes)")
	icecastClientTimeout = flag.Int("icecast-client-timeout", 30, "Icecast client timeout (s)")
	icecastHeaderTimeout = flag.Int("icecast-header-timeout", 15, "Icecast header timeout (s)")
	icecastSourceTimeout = flag.Int("icecast-source-timeout", 60, "Icecast source timeout (s)")
	icecastBurstSize     = flag.Int("icecast-burst-size", 131072, "Icecast connection burst size (bytes)")
)

// Return the global Icecast admin password. If it does not exist,
// generate a random one and save it for subsequent invocations.
func getIcecastAdminPassword() string {
	if icecastAdminPw == "" {
		data, err := ioutil.ReadFile(icecastAdminPwFile)
		if err != nil {
			icecastAdminPw = autoradio.GeneratePassword()
			if err := ioutil.WriteFile(icecastAdminPwFile, []byte(icecastAdminPw), 0600); err != nil {
				log.Printf("Couldn't save icecast admin password: %v", err)
			}
		} else {
			icecastAdminPw = string(data)
		}
	}
	return icecastAdminPw
}

type iceLimitsConfig struct {
	Clients       int `xml:"clients"`
	Sources       int `xml:"sources"`
	QueueSize     int `xml:"queue-size"`
	ClientTimeout int `xml:"client-timeout"`
	HeaderTimeout int `xml:"header-timeout"`
	SourceTimeout int `xml:"source-timeout"`
	BurstSize     int `xml:"burst-size"`
}

type iceAuthenticationConfig struct {
	SourcePassword string `xml:"source-password"`
	AdminUser      string `xml:"admin-user"`
	AdminPassword  string `xml:"admin-password"`
}

type iceListenConfig struct {
	BindAddress     string `xml:"bind-address"`
	Port            int    `xml:"port"`
	ShoutcastCompat int    `xml:"shoutcast-compat"`
}

type icePathsConfig struct {
	Basedir   string `xml:"basedir"`
	Logdir    string `xml:"logdir"`
	Webroot   string `xml:"webroot"`
	Adminroot string `xml:"adminroot"`
}

type iceLoggingConfig struct {
	Accesslog string `xml:"accesslog"`
	Errorlog  string `xml:"errorlog"`
	Loglevel  int    `xml:"loglevel"`
	Logsize   int    `xml:"logsize"`
}

type iceSecurityConfig struct {
	Chroot int `xml:"chroot"`
}

type iceRelayConfig struct {
	Server                 string `xml:"server"`
	Port                   int    `xml:"port"`
	Mount                  string `xml:"mount"`
	LocalMount             string `xml:"local-mount"`
	Username               string `xml:"username"`
	Password               string `xml:"password"`
	OnDemand               int    `xml:"on-demand"`
	RelayShoutcastMetadata int    `xml:"relay-shoutcast-metadata"`
}

type iceMountConfig struct {
	Name             string `xml:"mount-name"`
	Username         string `xml:"username"`
	Password         string `xml:"password"`
	FallbackMount    string `xml:"fallback-mount,omitempty"`
	FallbackOverride int    `xml:"fallback-override,omitempty"`
	Hidden           int    `xml:"hidden"`
	OnConnect        string `xml:"on-connect,omitempty"`
	OnDisconnect     string `xml:"on-disconnect,omitempty"`
	// Public           int    `xml:"no-yp"`
	// MaxListeners     int    `xml:"max-listeners"`
}

// Configuration of the local Icecast daemon. This is a write-only
// object, meant for serialization to XML. We keep around a single
// copy of it and just update Relays and Mounts every time.
type icecastConfig struct {
	XMLName   xml.Name
	Limits    iceLimitsConfig         `xml:"limits"`
	Auth      iceAuthenticationConfig `xml:"authentication"`
	Hostname  string                  `xml:"hostname"`
	Fileserve int                     `xml:"fileserve"`
	Paths     icePathsConfig          `xml:"paths"`
	Logging   iceLoggingConfig        `xml:"logging"`
	Security  iceSecurityConfig       `xml:"security"`
	Listen    []iceListenConfig       `xml:"listen-socket"`
	Relays    []iceRelayConfig        `xml:"relay"`
	Mounts    []iceMountConfig        `xml:"mount"`
}

// Create an Icecast configuration suitable for a Debian-based system
// install of the 'icecast2' package. Things to note about the
// generated config:
//
// - It binds to the IcecastPort (defined in api.go) on all
// interfaces.
//
// - A random admin password is generated once on each node, and saved
// to a file for persistence. It is not really meant to be used by the
// operator.
//
func newIcecastConfig(publicIp string, maxClients int) *icecastConfig {
	// We don't use the global source password, but icecast is
	// happier if it's set, so we just use a random password every
	// time.
	sourcePw := autoradio.GeneratePassword()
	adminPw := getIcecastAdminPassword()

	return &icecastConfig{
		XMLName: xml.Name{Local: "icecast"},
		Limits: iceLimitsConfig{
			Clients:       maxClients,
			Sources:       maxClients / 2,
			QueueSize:     *icecastQueueSize,
			ClientTimeout: *icecastClientTimeout,
			HeaderTimeout: *icecastHeaderTimeout,
			SourceTimeout: *icecastSourceTimeout,
			BurstSize:     *icecastBurstSize,
		},
		Auth: iceAuthenticationConfig{
			SourcePassword: sourcePw,
			AdminUser:      "admin",
			AdminPassword:  adminPw,
		},
		Hostname:  publicIp,
		Fileserve: 1,
		Paths: icePathsConfig{
			Basedir:   "/usr/share/icecast2",
			Logdir:    "/var/log/icecast2",
			Webroot:   "/usr/share/icecast2/web",
			Adminroot: "/usr/share/icecast2/admin",
		},
		Logging: iceLoggingConfig{
			Accesslog: "access.log",
			Errorlog:  "error.log",
			Loglevel:  3,
			Logsize:   10000,
		},
		Security: iceSecurityConfig{0},
		Listen: []iceListenConfig{
			{"0.0.0.0", autoradio.IcecastPort, 0},
		},
	}
}

// Encode the configuration to XML.
func (c *icecastConfig) Encode() ([]byte, error) {
	var buf bytes.Buffer

	output, err := xml.MarshalIndent(c, "", "  ")
	if err != nil {
		return nil, err
	}
	io.WriteString(&buf, "<!-- Automatically generated, do not edit -->\n\n")
	if _, err := buf.Write(output); err != nil {
		return nil, err
	}
	return buf.Bytes(), nil
}

// EncodeToFile writes the configuration to a file.
func (c *icecastConfig) EncodeToFile(path string) error {
	file, err := os.Create(path)
	if err != nil {
		return err
	}
	defer file.Close()

	data, err := c.Encode()
	if err != nil {
		return err
	}
	_, err = file.Write(data)
	return err
}

func masterMountToIcecastConfig(m *autoradio.Mount) iceMountConfig {
	mconfig := iceMountConfig{
		Name:     autoradio.MountNameToIcecastPath(m.Name),
		Username: m.Username,
		Password: m.Password,
		Hidden:   0,
		// MaxListeners: 1000,
		// NoYp:   1,
	}
	if m.Fallback != "" {
		mconfig.FallbackMount = m.Fallback
		mconfig.FallbackOverride = 1
	}
	return mconfig
}

func relayToIcecastConfig(m *autoradio.Mount) (iceRelayConfig, bool) {
	u, err := url.Parse(m.RelayUrl)
	if err != nil {
		// A failure here is almost invisible and not very
		// useful, but at least we can prevent garbling the
		// resulting icecast config.
		log.Printf("can't parse relay url for %s: %s", m.Name, err)
		return iceRelayConfig{}, false
	}
	server, port, err := net.SplitHostPort(u.Host)
	if err != nil {
		server = u.Host
		port = "80"
	}
	iport, _ := strconv.Atoi(port)

	rc := iceRelayConfig{
		Mount:                  u.Path,
		LocalMount:             autoradio.MountNameToIcecastPath(m.Name),
		Server:                 server,
		Port:                   iport,
		OnDemand:               1,
		RelayShoutcastMetadata: 1,
	}
	if u.User != nil {
		rc.Username = u.User.Username()
		if p, ok := u.User.Password(); ok {
			rc.Password = p
		}
	}
	return rc, true
}

func slaveMountToIcecastConfig(masterAddr string, m *autoradio.Mount) iceRelayConfig {
	path := autoradio.MountNameToIcecastPath(m.Name)
	return iceRelayConfig{
		Mount:                  path,
		LocalMount:             path,
		Server:                 masterAddr,
		Port:                   autoradio.IcecastPort,
		Username:               m.Username,
		Password:               m.Password,
		OnDemand:               1,
		RelayShoutcastMetadata: 1,
	}
}

// Update the configuration with the current list of mounts and
// masterelection state. This will clear the Mounts and Relays fields
// and set them to new values.
func (ic *icecastConfig) Update(config *clusterConfig, isMaster bool, masterAddr string) {
	var mounts []iceMountConfig
	var relays []iceRelayConfig

	for _, m := range config.ListMounts() {
		switch {
		case m.IsRelay():
			if rc, ok := relayToIcecastConfig(m); ok {
				relays = append(relays, rc)
			}
		case isMaster:
			mounts = append(mounts, masterMountToIcecastConfig(m))
		default:
			relays = append(relays, slaveMountToIcecastConfig(masterAddr, m))
		}
	}

	ic.Mounts = mounts
	ic.Relays = relays
}
